// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {crate::api_client::ApiClient, crate::constants, anyhow::Result, reqwest::Method};

pub async fn with(api_client: &ApiClient, calendar_name: &str) -> Result<()> {
    let report_method: Method = Method::from_bytes(b"REPORT")?;
    let events_xml = api_client
        .rest(|client| async {
            let response = client
                .request(
                    report_method.clone(),
                    api_client.caldav_base_url().join(&format!(
                        "{}/",
                        calendar_name.to_lowercase().replace(" ", "-")
                    ))?,
                )
                .header("Prefer", "return-minimal")
                .header("Content-Type", "application/xml; charset=utf-8")
                .header("Depth", 1)
                .body(format!(
                    "<c:calendar-query xmlns:d=\"DAV:\" xmlns:c=\"urn:ietf:params:xml:ns:caldav\">
                <d:prop>
                    <d:getetag/>
                </d:prop>
                <c:filter>
                    <c:comp-filter name=\"VCALENDAR\">
                        <c:prop-filter name=\"PRODID\">
                            <c:text-match>{}</c:text-match>
                        </c:prop-filter>
                    </c:comp-filter>
                </c:filter>
            </c:calendar-query>
            ",
                    constants::CALENDAR_PROVIDER,
                ))
                .send()
                .await;
            Ok(response?)
        })
        .await?
        .text()
        .await?;

    let xml_doc = libxml::parser::Parser::default().parse_string(events_xml.as_bytes())?;
    let mut xpath_ctx = libxml::xpath::Context::new(&xml_doc).unwrap();
    xpath_ctx.register_namespace("d", "DAV:").unwrap();
    let events_to_purge = xpath_ctx
        .findnodes("//d:response/d:href/text()", None)
        .unwrap()
        .into_iter()
        .map(|n| icalendar::parser::unfold(&n.get_content()))
        .collect::<Vec<_>>();

    for url in events_to_purge {
        api_client
            .rest(|client| async {
                Ok(client
                    .delete(api_client.base_url().join(&url)?)
                    .send()
                    .await?)
            })
            .await?;
        log::debug!("Purged {}", &url);
    }

    Ok(())
}
