// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

use {
    crate::recipe::Recipe,
    chrono::{NaiveDate, NaiveDateTime, NaiveTime, Utc},
    icalendar::Event,
    std::rc::Rc,
};

pub struct Scheduling {
    pub uid: String,
    pub ends_at: NaiveDateTime,
    pub recipe: Rc<Recipe>,
}

#[derive(strum_macros::Display)]
pub enum Meal {
    Lunch,
    Dinner,
}

impl Scheduling {
    pub fn new(date: NaiveDate, meal: Meal, recipe: Rc<Recipe>) -> Self {
        let uid = format!(
            "{}-{}@{}.montecristosoftware.eu",
            date,
            meal,
            env!("CARGO_PKG_NAME")
        );

        let meal_time = match meal {
            Meal::Lunch => NaiveTime::from_hms(12, 00, 00),
            Meal::Dinner => NaiveTime::from_hms(19, 00, 00),
        };

        let ends_at = NaiveDateTime::new(date, meal_time);

        Self {
            uid,
            ends_at,
            recipe,
        }
    }
}

impl From<Scheduling> for Event {
    fn from(ev: Scheduling) -> Self {
        use icalendar::Component;
        let start_time = ev.ends_at - ev.recipe.total_time();

        let mut cal_event = Event::new();

        cal_event
            .uid(&ev.uid)
            .summary(&ev.recipe.name)
            .description(&format!("cookbook@{}", ev.recipe.id))
            .timestamp(Utc::now())
            .starts(start_time)
            .ends(ev.ends_at);

        if let Some(ref location) = ev.recipe.url.clone() {
            cal_event.location(&location);
        }

        cal_event.done()
    }
}
